/*******************************************************************************
 * Copyright (c) 2011 Martin Schwengerer.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Martin Schwengerer - initial API and implementation
 ******************************************************************************/
package aspmodificator.base;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;

import main.RuleApplicationEnum;

import aspmodificator.model.LineBase;
import aspmodificator.modificationrules.IModificationRule;
import aspmodificator.modificationrules.TabuList;
import aspmodificator.modificationrules.gringo.DeleteRuleModificationRule;
import aspmodificator.parser.gringo.GringoRule;

public class ProgramModificator {

	protected Random random;
	private long seed;
	
	private boolean writeComments = true;
	private RuleApplicationEnum ruleApplication;
	private boolean useMemory = true;
	private TabuList tabuList;

	public ProgramModificator() {
		seed = System.currentTimeMillis();
		random = new Random(seed);
		tabuList = new TabuList();
	}

	private List<LineBase> copyProgram(List<LineBase> parsedProgram)
			throws CloneNotSupportedException {
		List<LineBase> result = new ArrayList<LineBase>();

		for (LineBase line : parsedProgram) {
			if (GringoRule.class == line.getClass())
				result.add((LineBase) line.clone());
			else
				result.add(line);
		}
		return result;
	}

	private int doModifications(List<IModificationRule> modificationRules,
			List<LineBase> copy) throws Exception {
		int modCount = 0;
		switch (ruleApplication) {
		case oneOfMany:
			Collections.shuffle(modificationRules, random);
			for (IModificationRule modificationRule : modificationRules) {
				modCount = modificationRule.apply(copy);

				if (modCount > 0)
					return modCount;
			}
			break;
		case all:
		default:
			// sort rules - let "delete rule-rules be first"
			sortRules(modificationRules);
			for (IModificationRule modificationRule : modificationRules) {
				modCount += modificationRule.apply(copy);
			}
			break;
		}

		return modCount;
	}

	public long getSeed() {
		return seed;
	}

	public List<LineBase> modifyProgram(List<LineBase> parsedProgram,
			List<IModificationRule> modificationRules) throws Exception {

		if (modificationRules == null)
			throw new NullPointerException("modificationRules must not be null");

		// create a copy of the parsed Program
		List<LineBase> copy = copyProgram(parsedProgram);

		// pass random to modification rules
		for (IModificationRule modificationRule : modificationRules) {
			modificationRule.setRandom(random);
			modificationRule.setWriteComments(writeComments);
			modificationRule.setUseMemory(useMemory);
			modificationRule.setTabuList(tabuList);
		}

		int noModifications = doModifications(modificationRules, copy);

		if (noModifications == 0)
			return null;

		return copy;
	}

	public void setRandomInitialization(long seed) {
		random.setSeed(seed);
	}

	public void setRuleApplication(RuleApplicationEnum ruleApplication) {
		this.ruleApplication = ruleApplication;
	}

	public void setUseMemory(boolean useMemory) {
		this.useMemory = useMemory;
	}

	public void setWriteComments(boolean writeComments) {
		this.writeComments = writeComments;
	}

	private void sortRules(List<IModificationRule> modificationRules) {
		List<IModificationRule> toInsertFirstList = new ArrayList<IModificationRule>();

		for (IModificationRule rule : modificationRules) {
			if (rule instanceof DeleteRuleModificationRule) {
				toInsertFirstList.add(rule);
			}
		}

		modificationRules.removeAll(toInsertFirstList);

		for (IModificationRule rule : toInsertFirstList) {
			modificationRules.add(0, rule);
		}
	}
}
