/*******************************************************************************
 * Copyright (c) 2011 Martin Schwengerer.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Martin Schwengerer - initial API and implementation
 ******************************************************************************/
package aspmodificator.modificationrules.gringo;

import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import aspmodificator.parser.gringo.GringoRule;
import aspmodificator.parser.gringo.Literal;
import aspmodificator.parser.gringo.MathTerm;
import aspmodificator.parser.gringo.Predicate;
import aspmodificator.parser.gringo.RelLit;
import aspmodificator.parser.gringo.VarTerm;

public class GringoRuleUtil {
	public static boolean checkSafety(GringoRule rule) {
		Set<String> headVariables = getVariables(rule.getHead());
		Set<String> negBodyVariables = getNegBodyVariables(rule.getBody());
		Set<String> otherNotSafeVariables = getOtherNotSafeVariables(rule
				.getBody());

		Set<String> varSet = new HashSet<String>();
		varSet.addAll(headVariables);
		varSet.addAll(negBodyVariables);
		varSet.addAll(otherNotSafeVariables);

		if (varSet.size() == 0)
			return true;

		if (rule.getBody() != null) {
			Iterator<Literal> it = rule.getBody().iterator();

			while (it.hasNext() && !varSet.isEmpty()) {
				Literal lit = it.next();
				removeBoundedVariables(varSet, lit);
			}

			if (varSet.isEmpty())
				return true;
		}

		return false;
	}

	private static Set<String> getNegBodyVariables(List<Literal> termList) {
		Set<String> result = new HashSet<String>();

		if (termList == null || termList.size() == 0)
			return result;

		for (Literal lit : termList) {
			if (!lit.isSigned())
				continue;

			List<VarTerm> vars = lit.getVariables();

			if (vars != null && vars.size() > 0) {
				for (VarTerm var : vars) {
					result.add(var.getIdentifier());
				}
			}
		}
		return result;
	}

	private static Set<String> getOtherNotSafeVariables(List<Literal> termList) {
		Set<String> result = new HashSet<String>();

		if (termList == null || termList.size() == 0)
			return result;

		for (Literal lit : termList) {

			// math terms
			List<Literal> literalList = Literal.findClassesByType(
					MathTerm.class, lit);
			for (Literal lit2 : literalList) {
				List<VarTerm> vars = lit2.getVariables();

				if (vars != null && vars.size() > 0) {
					for (VarTerm var : vars) {
						result.add(var.getIdentifier());
					}
				}
			}

			// relation literals
			literalList = Literal.findClassesByType(RelLit.class, lit);
			for (Literal lit2 : literalList) {
				List<VarTerm> vars = lit2.getVariables();

				if (vars != null && vars.size() > 0) {
					for (VarTerm var : vars) {
						result.add(var.getIdentifier());
					}
				}
			}
		}
		return result;
	}

	private static Set<String> getVariables(List<Literal> termList) {
		Set<String> result = new HashSet<String>();

		if (termList == null || termList.size() == 0)
			return result;

		for (Literal lit : termList) {
			List<VarTerm> vars = lit.getVariables();
			if (vars != null && vars.size() > 0) {
				for (VarTerm var : vars) {
					result.add(var.getIdentifier());
				}
			}
		}

		return result;
	}

	private static void removeBoundedVariables(Set<String> varSet, Literal lit) {
		// TODO Auto-generated method stub
		if (lit.isSigned())
			return;

		if (RelLit.class.equals(lit.getClass())
				&& (lit.getIdentifier().equals(RelLit.ASSIGN) || lit
						.getIdentifier().equals(RelLit.CASSIGN))) {
			if (lit.getTerms() != null && lit.getTerms().size() > 0) {
				List<VarTerm> varList = lit.getTerms().get(0).getVariables();
				if (varList != null && varList.size() == 1) {
					VarTerm saveVarCandidate = varList.get(0);

					// check special case when the variable on the left side
					// also occurs on the right side - if so, the variable is
					// not safe
					List<VarTerm> rightSideVarList = lit.getTerms().get(1)
							.getVariables();
					if (rightSideVarList != null && rightSideVarList.size() > 0) {

						for (VarTerm var : rightSideVarList) {
							if (var.getIdentifier().equals(
									saveVarCandidate.getIdentifier())) {
								return;
							}
						}
					}

					varSet.remove(saveVarCandidate.getIdentifier());
				}
			}

			return;
		}

		// if the literal is not a predicate, it can not bound variables -
		// do nothing
		if (!Predicate.class.isAssignableFrom(lit.getClass()))
			return;

		List<VarTerm> vars = lit.getVariables();

		if (vars != null && vars.size() > 0) {
			for (VarTerm var : vars) {
				varSet.remove(var.getIdentifier());
			}
		}

	}

}
