/*******************************************************************************
 * Copyright (c) 2011 Martin Schwengerer.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Martin Schwengerer - initial API and implementation
 ******************************************************************************/
package aspmodificator.modificationrules.gringo;

import java.util.ArrayList;
import java.util.List;

import aspmodificator.model.LineBase;
import aspmodificator.parser.gringo.GringoRule;
import aspmodificator.parser.gringo.Literal;
import aspmodificator.parser.gringo.MathTerm;

public class MathOperatorsModificationRule extends AbstractModificationRule {

	public static final String[] dualOperators = { MathTerm.DIV, MathTerm.MOD,
			MathTerm.PLUS, MathTerm.MINUS, MathTerm.POW, MathTerm.MULT };
	public static final String[] singleOperators = { MathTerm.AND,
			MathTerm.XOR, MathTerm.BNOT };

	public MathOperatorsModificationRule() {
		super();
	}

	public MathOperatorsModificationRule(int minApplication, int maxApplication) {
		super(minApplication, maxApplication);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * aspmodificator.modificationrules.IModificationRule#apply(java.util.List)
	 */
	@Override
	public int apply(List<LineBase> program) {
		// find applicable rules/objects
		List<Literal> applicableObject = findApplicableObjects(program);

		if (0 == applicableObject.size())
			return 0;
		int ruleApplied = 0;
		double chance = (double) (minApplication + maxApplication)
				/ (double) (2 * applicableObject.size());

		int applicableObjectNo = applicableObject.size();

		while (applicableObjectNo > 0 && ruleApplied < maxApplication
				&& ruleApplied < minApplication) {

			applicableObjectNo = applicableObject.size();
			int objectIndex = -1;

			for (Literal lit : applicableObject) {

				objectIndex++;

				if (lit.isModified()) {
					applicableObjectNo--;
					continue;
				}

				double randomValue = random.nextDouble();

				if (randomValue <= chance) {

					int index = -1;
					if (lit.getTerms().size() == 1) {
						index = random.nextInt(singleOperators.length);
						int tries = 0;
						int maxTries = (singleOperators.length - 1) * 2;
						while (singleOperators[index].equals(lit
								.getIdentifier())
								|| (useMemory
										&& tabuList.isTabu(getRuleTypeEnum(),
												lit.getLineNo(), objectIndex,
												index) && tries < maxTries)) {

							index = random.nextInt(singleOperators.length);
							tries++;
						}

						if (useMemory && tries >= maxTries) {
							applicableObjectNo--;
							continue;
						}

						modifyLiteral(lit, singleOperators[index]);
					} else {
						int tries = 0;
						int maxTries = (dualOperators.length - 1) * 2;

						index = random.nextInt(dualOperators.length);
						while (dualOperators[index].equals(lit.getIdentifier())
								|| (useMemory
										&& tabuList.isTabu(getRuleTypeEnum(),
												lit.getLineNo(), objectIndex,
												index) && tries < maxTries)) {
							index = random.nextInt(dualOperators.length);
							tries++;
						}

						if (useMemory && tries >= maxTries) {
							applicableObjectNo--;
							continue;
						}

						modifyLiteral(lit, dualOperators[index]);
					}

					lit.setModified(true);

					ruleApplied++;

					applicableObjectNo--;

					if (useMemory)
						tabuList.setTabu(getRuleTypeEnum(), lit.getLineNo(),
								objectIndex, index);

					if (ruleApplied == maxApplication)
						break;
				}
			}
		}

		return ruleApplied;
	}

	protected List<Literal> findApplicableObjects(List<LineBase> program) {
		List<Literal> result = new ArrayList<Literal>();

		for (LineBase line : program) {
			if (GringoRule.class.isAssignableFrom(line.getClass())) {
				GringoRule rule = (GringoRule) line;

				if (rule.getHead() != null
						&& (mrRuleTarget == RuleTargetEnum.ALL || mrRuleTarget == RuleTargetEnum.HEAD)) {
					for (Literal lit : rule.getHead()) {
						if (lit.isModified())
							continue;

						List<Literal> tmpList = Literal.findClassesByType(
								MathTerm.class, lit);

						for (Literal mt : tmpList) {
							if (!mt.isModified()) {
								if (mt.getTerms() == null
										|| mt.getTerms().size() == 0
										|| mt.getTerms().size() > 2)
									continue;

								result.add(mt);
							}
						}
					}
				}

				if (rule.getBody() != null
						&& (mrRuleTarget == RuleTargetEnum.ALL || mrRuleTarget == RuleTargetEnum.BODY)) {
					for (Literal lit : rule.getBody()) {
						if (lit.isModified())
							continue;

						List<Literal> tmpList = Literal.findClassesByType(
								MathTerm.class, lit);

						for (Literal mt : tmpList) {
							if (!mt.isModified())
								result.add(mt);
						}
					}
				}
			}
		}

		return result;
	}

	@Override
	public ModificationRuleTypeEnum getRuleTypeEnum() {
		return ModificationRuleTypeEnum.ARO;
	}

	protected void modifyLiteral(Literal lit, String newIdentifier) {

		if (writeComments) {
			lit.addSucceedingComment("%* ModificationNote: operator has been modified. Original operator was \""
					+ lit.getIdentifier() + "\" *%");
		}

		lit.setIdentifier(newIdentifier);
	}

}
