/*******************************************************************************
 * Copyright (c) 2011 Martin Schwengerer.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Martin Schwengerer - initial API and implementation
 ******************************************************************************/
package aspmodificator.modificationrules.gringo;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import aspmodificator.model.LineBase;
import aspmodificator.parser.gringo.GringoRule;
import aspmodificator.parser.gringo.Literal;
import aspmodificator.parser.gringo.Predicate;

public class RenamePredicateModificationRule extends AbstractModificationRule {

	public RenamePredicateModificationRule() {
		super();
	}

	public RenamePredicateModificationRule(int minApplication,
			int maxApplication) {
		super(minApplication, maxApplication);
	}

	@Override
	public int apply(List<LineBase> program) {
		HashMap<Literal, List<Literal>> substitutions = findEquivalentSubstitutions(
				program, Predicate.class);

		if (substitutions == null || substitutions.size() == 0)
			return 0;

		// find applicable rules/objects
		List<Literal> applicableObject = findApplicableObjects(program,
				substitutions.keySet());

		if (0 == applicableObject.size())
			return 0;
		int ruleApplied = 0;
		double chance = (double) (minApplication + maxApplication)
				/ (double) (2 * applicableObject.size());

		int applicableObjectNo = applicableObject.size();

		while (applicableObjectNo > 0 && ruleApplied < maxApplication
				&& ruleApplied < minApplication) {

			applicableObjectNo = applicableObject.size();
			int objectIndex = -1;

			for (Literal lit : applicableObject) {

				objectIndex++;

				if (lit.isModified()) {
					applicableObjectNo--;
					continue;
				}

				if (useMemory
						&& tabuList.isTabu(getRuleTypeEnum(), lit.getLineNo(),
								objectIndex)) {
					applicableObjectNo--;
					continue;
				}

				double randomValue = random.nextDouble();

				if (randomValue <= chance) {
					modifyLiteral(lit, substitutions);
					lit.setModified(true);

					ruleApplied++;

					applicableObjectNo--;

					if (useMemory)
						tabuList.setTabu(getRuleTypeEnum(), lit.getLineNo(),
								objectIndex);

					if (ruleApplied == maxApplication)
						break;
				}
			}
		}

		return ruleApplied;
	}

	private boolean contains(Set<Literal> literalSet, Literal l) {
		for (Literal lit : literalSet) {
			if (lit.equalSignature(l))
				return true;
		}
		return false;
	}

	private List<Literal> findApplicableObjects(List<LineBase> program,
			Set<Literal> keySet) {
		List<Literal> result = new ArrayList<Literal>();
		for (LineBase line : program) {
			if (GringoRule.class.isAssignableFrom(line.getClass())) {
				GringoRule rule = (GringoRule) line;

				if (rule.getHead() != null
						&& (mrRuleTarget == RuleTargetEnum.ALL || mrRuleTarget == RuleTargetEnum.HEAD)) {
					for (Literal lit : rule.getHead()) {
						if (contains(keySet, lit))
							result.add(lit);
					}
				}

				if (rule.getBody() != null
						&& (mrRuleTarget == RuleTargetEnum.ALL || mrRuleTarget == RuleTargetEnum.BODY)) {
					for (Literal lit : rule.getBody()) {
						if (contains(keySet, lit))
							result.add(lit);
					}
				}
			}
		}
		return result;
	}

	private HashMap<Literal, List<Literal>> findEquivalentSubstitutions(
			List<LineBase> program, Class<?> clazz) {
		HashMap<Literal, List<Literal>> result = new HashMap<Literal, List<Literal>>();

		for (LineBase line : program) {
			if (GringoRule.class.isAssignableFrom(line.getClass())) {
				GringoRule rule = (GringoRule) line;

				if (rule.getHead() != null) {
					for (Literal lit : rule.getHead()) {
						List<Literal> tmpList = Literal.findClassesByType(
								clazz, lit);

						for (Literal l : tmpList) {
							if (!contains(result.keySet(), l))
								result.put(l, new ArrayList<Literal>());
						}
					}
				}

				if (rule.getBody() != null) {
					for (Literal lit : rule.getBody()) {
						List<Literal> tmpList = Literal.findClassesByType(
								clazz, lit);

						for (Literal l : tmpList) {
							if (!contains(result.keySet(), l))
								result.put(l, new ArrayList<Literal>());
						}
					}
				}
			}
		}

		for (Literal lit : result.keySet()) {
			for (Literal lit2 : result.keySet()) {
				if (lit == lit2)
					continue;

				if (lit.getArity() == lit2.getArity())
					result.get(lit).add(lit2);
			}
		}

		List<Literal> toRemove = new LinkedList<Literal>();
		for (Literal lit : result.keySet()) {
			if (result.get(lit).size() == 0) {
				toRemove.add(lit);
			} else {
				// this sorting is just needed because it seems that the order
				// of the list-members is not deterministic
				Collections.sort(result.get(lit), new Comparator<Literal>() {
					@Override
					public int compare(Literal arg0, Literal arg1) {
						if (arg0.getIdentifier() == null)
							return -1;
						if (arg1.getIdentifier() == null)
							return 1;

						return arg0.getIdentifier().compareTo(
								arg1.getIdentifier());
					}

				});
			}
		}

		for (Literal lit : toRemove)
			result.remove(lit);

		return result;
	}

	@Override
	public ModificationRuleTypeEnum getRuleTypeEnum() {
		return ModificationRuleTypeEnum.LRP;
	}

	private void modifyLiteral(Literal lit,
			HashMap<Literal, List<Literal>> substitutions) {
		List<Literal> possSubstitutions = null;
		for (Literal l : substitutions.keySet()) {
			if (l.equalSignature(lit)) {
				possSubstitutions = substitutions.get(l);
				break;
			}
		}

		if (possSubstitutions == null || possSubstitutions.size() == 0)
			return;

		if (writeComments) {
			lit.addSucceedingComment("%*ModificationNote: predicate has been renamed. Original name was "
					+ lit.getIdentifier() + " *%");
		}

		int index = random.nextInt(possSubstitutions.size());

		lit.setIdentifier(possSubstitutions.get(index).getIdentifier());

	}

}
