/*******************************************************************************
 * Copyright (c) 2011 Martin Schwengerer.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Martin Schwengerer - initial API and implementation
 ******************************************************************************/
package aspmodificator.modificationrules.gringo;

import java.util.ArrayList;
import java.util.List;

import aspmodificator.model.LineBase;
import aspmodificator.parser.gringo.ConstTerm;
import aspmodificator.parser.gringo.GringoRule;
import aspmodificator.parser.gringo.Literal;
import aspmodificator.parser.gringo.MathTerm;
import aspmodificator.parser.gringo.RelLit;

public class TwiddleMathStatementModificationRule extends
		AbstractModificationRule {

	public boolean includeMathTerms = true;
	public boolean includeRelLiterals = true;

	public static final String COMMENT_TWIDDLEMATH_INC = "%*ModificationNote: +1 added *%";
	public static final String COMMENT_TWIDDLEMATH_DEC = "%*ModificationNote: -1 added *%";

	public TwiddleMathStatementModificationRule() {
		super();
	}

	public TwiddleMathStatementModificationRule(int min, int max) {
		super(min, max);
	}

	public TwiddleMathStatementModificationRule(int minApplication,
			int maxApplication, boolean includeRelations,
			boolean includeMathTerms) {
		this(minApplication, maxApplication);
		this.includeMathTerms = includeMathTerms;
		this.includeRelLiterals = includeRelations;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * aspmodificator.modificationrules.IModificationRule#apply(java.util.List)
	 */
	@Override
	public int apply(List<LineBase> program) {
		// find applicable rules/objects
		List<Literal> applicableObject = findApplicableObjects(program);

		if (0 == applicableObject.size())
			return 0;
		int ruleApplied = 0;
		double chance = (double) (minApplication + maxApplication)
				/ (double) (2 * applicableObject.size());

		int applicableObjectNo = applicableObject.size();

		while (applicableObjectNo > 0 && ruleApplied < maxApplication
				&& ruleApplied < minApplication) {

			applicableObjectNo = applicableObject.size();
			int objectIndex = -1;

			for (Literal lit : applicableObject) {

				objectIndex++;

				if (lit.isModified()) {
					applicableObjectNo--;
					continue;
				}

				double randomValue = random.nextDouble();

				if (randomValue <= chance) {
					int termSize = lit.getTerms().size();

					int choosenIndex = random.nextInt(termSize);
					boolean shouldIncrement = random.nextBoolean();

					// special case to prevent X =Y+1 -> X+1=Y+1
					choosenIndex = preventIllegalAssignment(lit, choosenIndex);

					if (useMemory && shouldIncrement)
						choosenIndex += termSize;

					int tries = 0;
					int maxTries = termSize * 2;
					while (useMemory
							&& tabuList.isTabu(getRuleTypeEnum(),
									lit.getLineNo(), objectIndex, choosenIndex)
							&& tries < maxTries) {
						choosenIndex = random.nextInt(termSize);
						shouldIncrement = random.nextBoolean();

						// special case to prevent X =Y+1 -> X+1=Y+1
						choosenIndex = preventIllegalAssignment(lit,
								choosenIndex);

						if (shouldIncrement)
							choosenIndex += termSize;

						tries++;
					}

					if (useMemory
							&& tabuList.isTabu(getRuleTypeEnum(),
									lit.getLineNo(), objectIndex, choosenIndex)) {
						applicableObjectNo--;
						continue;
					}

					if (useMemory && shouldIncrement)
						choosenIndex -= termSize;

					modifyLiteral(lit, choosenIndex, shouldIncrement);
					lit.setModified(true);

					ruleApplied++;

					applicableObjectNo--;

					if (useMemory) {
						if (shouldIncrement)
							choosenIndex += termSize;
						tabuList.setTabu(getRuleTypeEnum(), lit.getLineNo(),
								objectIndex, choosenIndex);
					}

					if (ruleApplied == maxApplication)
						break;
				}
			}
		}

		return ruleApplied;
	}

	protected List<Literal> findApplicableObjects(List<LineBase> program) {
		List<Literal> result = new ArrayList<Literal>();

		for (LineBase line : program) {
			if (GringoRule.class.isAssignableFrom(line.getClass())) {
				GringoRule rule = (GringoRule) line;

				if (rule.getHead() != null
						&& (mrRuleTarget == RuleTargetEnum.ALL || mrRuleTarget == RuleTargetEnum.HEAD)) {
					for (Literal lit : rule.getHead()) {
						if (lit.isModified())
							continue;

						// find all MathTerms
						List<Literal> tmpList;

						if (includeMathTerms) {
							tmpList = Literal.findClassesByType(MathTerm.class,
									lit);

							for (Literal mt : tmpList) {
								if (!mt.isModified())
									result.add(mt);
							}
						}

						// find all RelLits
						if (includeRelLiterals) {
							tmpList = Literal.findClassesByType(RelLit.class,
									lit);

							for (Literal mt : tmpList) {
								if (!mt.isModified())
									result.add(mt);
							}
						}
					}
				}

				if (rule.getBody() != null
						&& (mrRuleTarget == RuleTargetEnum.ALL || mrRuleTarget == RuleTargetEnum.BODY)) {
					for (Literal lit : rule.getBody()) {
						if (lit.isModified())
							continue;

						// find all MathTerms
						List<Literal> tmpList;

						if (includeMathTerms) {
							tmpList = Literal.findClassesByType(

							MathTerm.class, lit);

							for (Literal mt : tmpList) {
								if (!mt.isModified())
									result.add(mt);
							}
						}

						// find all RelLits
						if (includeRelLiterals) {
							tmpList = Literal.findClassesByType(RelLit.class,
									lit);

							for (Literal mt : tmpList) {
								if (!mt.isModified())
									result.add(mt);
							}
						}
					}
				}
			}
		}

		return result;
	}

	@Override
	public ModificationRuleTypeEnum getRuleTypeEnum() {
		return ModificationRuleTypeEnum.ATV;
	}

	public boolean isIncludeMathTerms() {
		return includeMathTerms;
	}

	public boolean isIncludeRelLiterals() {
		return includeRelLiterals;
	}

	protected void modifyLiteral(Literal lit, int toChangeIndex,
			boolean shouldIncrement) {
		List<Literal> terms = lit.getTerms();

		Literal orig = terms.get(toChangeIndex);

		MathTerm mathTerm;

		if (shouldIncrement) {
			mathTerm = new MathTerm(MathTerm.PLUS, orig, new ConstTerm(
					Integer.valueOf(1), lit.getLineNo()), lit.getLineNo());

			if (writeComments)
				mathTerm.addPrecedingComment(COMMENT_TWIDDLEMATH_INC);
		} else {
			mathTerm = new MathTerm(MathTerm.MINUS, orig, new ConstTerm(
					Integer.valueOf(1), lit.getLineNo()), lit.getLineNo());
			if (writeComments)
				mathTerm.addPrecedingComment(COMMENT_TWIDDLEMATH_DEC);
		}

		terms.set(toChangeIndex, mathTerm);

	}

	private int preventIllegalAssignment(Literal lit, int choosenIndex) {
		if (lit.getIdentifier() == RelLit.ASSIGN
				|| lit.getIdentifier() == RelLit.CASSIGN) {
			if (lit.getTerms().get(choosenIndex).getArity() == 0
					&& lit.getTerms().get(1 - choosenIndex).getArity() == 2)
				choosenIndex = 1 - choosenIndex;
		}
		return choosenIndex;
	}

	public void setIncludeMathTerms(boolean includeMathTerms) {
		this.includeMathTerms = includeMathTerms;
	}

	public void setIncludeRelLiterals(boolean includeRelLiterals) {
		this.includeRelLiterals = includeRelLiterals;
	}

}
