/*******************************************************************************
 * Copyright (c) 2011 Martin Schwengerer.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Martin Schwengerer - initial API and implementation
 ******************************************************************************/
package aspmodificator.modificationrules;

import static org.hamcrest.core.Is.is;
import static org.junit.Assert.assertThat;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.util.List;
import java.util.Random;

import org.junit.Before;
import org.junit.Test;

import aspmodificator.model.LineBase;
import aspmodificator.modificationrules.gringo.DeleteTermModificationRule;
import aspmodificator.modificationrules.gringo.ModificationRuleTypeEnum;
import aspmodificator.modificationrules.gringo.RenameVariableModificationRule;
import aspmodificator.modificationrules.gringo.RuleTargetEnum;
import aspmodificator.parser.GringoParser;

public class TabuListTest {

	private TabuList tabuList;
	private DeleteTermModificationRule rule;
	private GringoParser parser;

	@Test
	public void isTabu_shouldReturnFalse_1() {
		assertThat(tabuList.isTabu(ModificationRuleTypeEnum.RDP, 1, 1),
				is(false));
	}

	@Test
	public void isTabu_shouldReturnFalse_2() {
		tabuList.setTabu(ModificationRuleTypeEnum.RDP, 2, 1);

		assertThat(tabuList.isTabu(ModificationRuleTypeEnum.RDP, 1, 1),
				is(false));
	}

	@Test
	public void isTabu_shouldReturnFalse_3() {
		tabuList.setTabu(ModificationRuleTypeEnum.RDP, 1, 3);

		assertThat(tabuList.isTabu(ModificationRuleTypeEnum.RDP, 1, 1),
				is(false));
	}

	@Test
	public void isTabu_shouldReturnFalse_4() {
		tabuList.setTabu(ModificationRuleTypeEnum.ARO, 1, 1);

		assertThat(tabuList.isTabu(ModificationRuleTypeEnum.RDP, 1, 1),
				is(false));
	}

	@Test
	public void isTabu_shouldReturnTrue_1() {
		tabuList.setTabu(ModificationRuleTypeEnum.RDP, 1, 1);

		assertThat(tabuList.isTabu(ModificationRuleTypeEnum.RDP, 1, 1),
				is(true));
	}

	@Test
	public void isTabu_shouldReturnTrue_2() {
		tabuList.setTabu(ModificationRuleTypeEnum.RDP, 1, 1);
		tabuList.setTabu(ModificationRuleTypeEnum.RDP, 1, 2);
		tabuList.setTabu(ModificationRuleTypeEnum.RDP, 1, 3);

		assertThat(tabuList.isTabu(ModificationRuleTypeEnum.RDP, 1, 1),
				is(true));
	}

	@Before
	public void setUp() throws Exception {
		tabuList = new TabuList();
		rule = new DeleteTermModificationRule();
		parser = new GringoParser();
	}

	@Test
	public void testTabuList_2Dim_shouldCreateDifferentPrograms_1()
			throws Exception {
		RenameVariableModificationRule renameVariableRule = new RenameVariableModificationRule();
		String code = "a(X,Y,Z) :- a(X), b(Y), c(Z), d(X,Y,Z).";
		List<LineBase> program = parser.parse(code);

		renameVariableRule.setWriteComments(false);
		renameVariableRule.setMinApplication(1);
		renameVariableRule.setMaxApplication(1);
		renameVariableRule.setMrRuleTarget(RuleTargetEnum.ALL);
		renameVariableRule.setTabuList(tabuList);
		renameVariableRule.setUseMemory(true);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0).thenReturn(0.0);
		when(mockedRandom.nextInt(2)).thenReturn(0).thenReturn(1);
		renameVariableRule.setRandom(mockedRandom);

		int res = renameVariableRule.apply(program);

		assertThat(res, is(1));
		assertThat(program.get(0).toString(),
				is("a(Y,Y,Z) :- a(X), b(Y), c(Z), d(X,Y,Z)."));

		program = parser.parse(code);

		renameVariableRule.apply(program);

		assertThat(res, is(1));
		assertThat(program.get(0).toString(),
				is("a(Z,Y,Z) :- a(X), b(Y), c(Z), d(X,Y,Z)."));

	}

	@Test
	public void testTabuList_2Dim_shouldCreateDifferentPrograms_2()
			throws Exception {
		RenameVariableModificationRule renameVariableRule = new RenameVariableModificationRule();
		String code = "a(X,Y) :- b(X), b(Y).";
		List<LineBase> program = parser.parse(code);

		renameVariableRule.setWriteComments(false);
		renameVariableRule.setMinApplication(1);
		renameVariableRule.setMaxApplication(1);
		renameVariableRule.setMrRuleTarget(RuleTargetEnum.ALL);
		renameVariableRule.setTabuList(tabuList);
		renameVariableRule.setUseMemory(true);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0).thenReturn(0.0);
		when(mockedRandom.nextInt(1)).thenReturn(0).thenReturn(0);
		renameVariableRule.setRandom(mockedRandom);

		int res = renameVariableRule.apply(program);

		assertThat(res, is(1));
		assertThat(program.get(0).toString(), is("a(Y,Y) :- b(X), b(Y)."));

		program = parser.parse(code);

		renameVariableRule.apply(program);

		assertThat(res, is(1));
		assertThat(program.get(0).toString(), is("a(X,X) :- b(X), b(Y)."));

	}

	@Test
	public void testTabuList_shouldCreateDifferentMutants_1() throws Exception {
		String code = "a(X,Y,Z) :- a(X), b(Y), c(Z), d(X,Y,Z).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setMrRuleTarget(RuleTargetEnum.BODY);
		rule.setTabuList(tabuList);
		rule.setUseMemory(true);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0).thenReturn(0.0)
				.thenReturn(0.0);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);

		assertThat(res, is(1));
		assertThat(program.get(0).toString(),
				is("a(X,Y,Z) :- b(Y), c(Z), d(X,Y,Z)."));

		program = parser.parse(code);
		rule.apply(program);

		assertThat(res, is(1));
		assertThat(program.get(0).toString(),
				is("a(X,Y,Z) :- a(X), c(Z), d(X,Y,Z)."));
	}

	@Test
	public void testTabuList_shouldNotCreateDifferentMutants_1()
			throws Exception {
		String code = "a(X,Y,Z) :- a(X), b(Y), c(Z), d(X,Y,Z).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setMrRuleTarget(RuleTargetEnum.BODY);
		rule.setTabuList(tabuList);
		rule.setUseMemory(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0).thenReturn(0.0)
				.thenReturn(0.0);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);

		assertThat(res, is(1));
		assertThat(program.get(0).toString(),
				is("a(X,Y,Z) :- b(Y), c(Z), d(X,Y,Z)."));

		program = parser.parse(code);
		rule.apply(program);

		assertThat(res, is(1));
		assertThat(program.get(0).toString(),
				is("a(X,Y,Z) :- b(Y), c(Z), d(X,Y,Z)."));
	}

}
