/*******************************************************************************
 * Copyright (c) 2011 Martin Schwengerer.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Martin Schwengerer - initial API and implementation
 ******************************************************************************/
package aspmodificator.modificationrules.gringo;

import static org.junit.Assert.*;

import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Random;

import org.junit.Before;
import org.junit.Test;

import aspmodificator.model.LineBase;
import aspmodificator.parser.GringoParser;
import aspmodificator.parser.gringo.AggregateLiteral;
import aspmodificator.parser.gringo.GringoRule;
import aspmodificator.parser.gringo.Literal;
import static org.hamcrest.CoreMatchers.*;
import static org.hamcrest.core.Is.is;

public class AggregateLiteralModificationRuleTest {

	private GringoParser parser;
	private AggregateLiteralModificationRule rule;

	@Before
	public void setUp() throws Exception {
		rule = new AggregateLiteralModificationRule();
		parser = new GringoParser();
	}

	@Test
	public void testApply_shouldModify_1() throws Exception {
		String code = "2 { enroll(C) : course(C,S,H) } 2.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextBoolean()).thenReturn(false).thenReturn(true);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertEquals(1, res);
		assertEquals("2 { enroll(C) : course(C,S,H) } 3",
				((GringoRule) program.get(0)).getHead().get(0).toString());
	}

	@Test
	public void testApply_shouldModify_2() throws Exception {
		String code = "2 { enroll(C) : course(C,S,H) } 2.\n"
				+ ":- [ enroll(C) : course(C,S,H) ] 10.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(1.0).thenReturn(0.0);
		when(mockedRandom.nextBoolean()).thenReturn(false).thenReturn(false);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertEquals(1, res);
		assertEquals("[ enroll(C) : course(C,S,H) ] 9",
				((GringoRule) program.get(1)).getBody().get(0).toString());
	}

	@Test
	public void testApply_shouldModify_3() throws Exception {
		String code = "1{guess(X)}N :- max(N).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(1.0).thenReturn(0.0);
		when(mockedRandom.nextBoolean()).thenReturn(false).thenReturn(false);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertThat(res, is(1));
		assertThat(((GringoRule) program.get(0)).toString(),
				is("1 { guess(X) } N-1 :- max(N)."));
	}

	@Test
	public void testApply_shouldNotModify_1() throws Exception {
		String code = "a(X) :- X = #min [ enroll(C) : hours(C,H) = H ].\n"
				+ ":-  X = #max [ enroll(C) : hours(C,H) = H ], X > 2.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(1.0).thenReturn(0.0);
		when(mockedRandom.nextBoolean()).thenReturn(false).thenReturn(false);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertThat(res, is(0));
	}

	@Test
	public void testFindApplicableObjects_shouldReturn1_1() throws Exception {
		String code = "2 { enroll(C) : course(C,S,H) } 2.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertThat(literalList, is(notNullValue()));
		assertThat(literalList.size(), is(1));
	}

	@Test
	public void testFindApplicableObjects_shouldReturn1_2() throws Exception {
		String code = ":- [ enroll(C) : course(C,S,H) ] 10.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertThat(literalList, is(notNullValue()));
		assertThat(literalList.size(), is(1));
	}

	@Test
	public void testFindApplicableObjects_shouldReturn1_3() throws Exception {
		String code = ":- 6 [ enroll(C) : course(C,3,H), enroll(C) : course(C,4,H) ].";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertThat(literalList, is(notNullValue()));
		assertThat(literalList.size(), is(1));
	}

	@Test
	public void testFindApplicableObjects_shouldReturn4() throws Exception {
		String code = ":- [ enroll(1) = 1, enroll(8) = 1 ] 10.\n"
				+ ":- [ enroll(C) : course(C,S,H) ] 10.\n"
				+ ":- 2 [ not enroll(C) : course(C,2,H) ].\n"
				+ ":- not M-2 [ enroll(C) : hours(C,H) = H ] M, max_hours(M).\n"
				+ ":- 6 #max [ enroll(C) : hours(C,H) = H ].";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(5, literalList.size());
		assertEquals("[ enroll(C) : course(C,S,H) ] 10", literalList.get(1)
				.toString());
		assertEquals("2 [ not enroll(C) : course(C,2,H) ]", literalList.get(2)
				.toString());
		assertEquals("M-2 [ enroll(C) : hours(C,H) = H ] M", literalList.get(3)
				.toString());
		assertEquals("6 #max [ enroll(C) : hours(C,H) = H ]", literalList
				.get(4).toString());
	}

	@Test
	public void testFindApplicableObjects_shouldReturnEmptyList_1()
			throws Exception {
		String code = "";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertThat(literalList, is(notNullValue()));
		assertThat(literalList.size(), is(0));
	}

	@Test
	public void testFindApplicableObjects_shouldReturnEmptyList_2()
			throws Exception {
		String code = "a(x). \n a(X) :- b(X). \n :- c(x).";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertThat(literalList, is(notNullValue()));
		assertThat(literalList.size(), is(0));
	}

	@Test
	public void testFindApplicableObjects_shouldReturnEmptyList_3()
			throws Exception {
		String code = "a(X) :- X = #min [ enroll(C) : hours(C,H) = H ].";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertThat(literalList, is(notNullValue()));
		assertThat(literalList.size(), is(0));
	}

	@Test
	public void testModifyLiteral_shouldDecrementLowerBound_1()
			throws Exception {
		String code = "4 { enroll(C) : course(C,S,H) } 6.";
		List<LineBase> program = parser.parse(code);
		AggregateLiteral literal = (AggregateLiteral) ((GringoRule) program
				.get(0)).getHead().get(0);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextBoolean()).thenReturn(true).thenReturn(false);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(literal);
		assertEquals("3 { enroll(C) : course(C,S,H) } 6", literal.toString());
	}

	@Test
	public void testModifyLiteral_shouldDecrementLowerBound_2()
			throws Exception {
		String code = "2 { enroll(C) : course(C,S,H) } 2.";
		List<LineBase> program = parser.parse(code);
		AggregateLiteral literal = (AggregateLiteral) ((GringoRule) program
				.get(0)).getHead().get(0);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextBoolean()).thenReturn(true).thenReturn(false);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(literal);
		assertEquals("1 { enroll(C) : course(C,S,H) } 2", literal.toString());
	}

	@Test
	public void testModifyLiteral_shouldDecrementLowerBound_3()
			throws Exception {
		String code = "1 { enroll(C) : course(C,S,H) } 1.";
		List<LineBase> program = parser.parse(code);
		AggregateLiteral literal = (AggregateLiteral) ((GringoRule) program
				.get(0)).getHead().get(0);
		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextBoolean()).thenReturn(true).thenReturn(false);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(literal);
		assertEquals("0 { enroll(C) : course(C,S,H) } 1", literal.toString());
	}

	@Test
	public void testModifyLiteral_shouldDecrementLowerBound_4()
			throws Exception {
		String code = "1 { enroll(C) : course(C,S,H) } 1.";
		List<LineBase> program = parser.parse(code);
		AggregateLiteral literal = (AggregateLiteral) ((GringoRule) program
				.get(0)).getHead().get(0);
		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextBoolean()).thenReturn(true).thenReturn(true);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(literal);
		assertEquals("0 { enroll(C) : course(C,S,H) } 1", literal.toString());
	}

	@Test
	public void testModifyLiteral_shouldDecrementUpperBound_1()
			throws Exception {
		String code = "3 { enroll(C) : course(C,S,H) } 7.";
		List<LineBase> program = parser.parse(code);
		AggregateLiteral literal = (AggregateLiteral) ((GringoRule) program
				.get(0)).getHead().get(0);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextBoolean()).thenReturn(false).thenReturn(false);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(literal);
		assertEquals("3 { enroll(C) : course(C,S,H) } 6", literal.toString());
	}

	@Test
	public void testModifyLiteral_shouldDecrementUpperBound_2()
			throws Exception {
		String code = "3 { enroll(C) : course(C,S,H) } 4.";
		List<LineBase> program = parser.parse(code);
		AggregateLiteral literal = (AggregateLiteral) ((GringoRule) program
				.get(0)).getHead().get(0);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextBoolean()).thenReturn(false).thenReturn(false);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(literal);
		assertEquals("3 { enroll(C) : course(C,S,H) } 3", literal.toString());
	}

	@Test
	public void testModifyLiteral_shouldDecrementUpperBound_3()
			throws Exception {
		String code = "1 { enroll(C) : course(C,S,H) } 1.";
		List<LineBase> program = parser.parse(code);
		AggregateLiteral literal = (AggregateLiteral) ((GringoRule) program
				.get(0)).getHead().get(0);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextBoolean()).thenReturn(false).thenReturn(false);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(literal);
		assertEquals("1 { enroll(C) : course(C,S,H) } 2", literal.toString());
	}

	@Test
	public void testModifyLiteral_shouldIncrementLowerBound_1()
			throws Exception {
		String code = "3 { enroll(C) : course(C,S,H) } 6.";
		List<LineBase> program = parser.parse(code);
		AggregateLiteral literal = (AggregateLiteral) ((GringoRule) program
				.get(0)).getHead().get(0);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextBoolean()).thenReturn(true).thenReturn(true);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(literal);
		assertEquals("4 { enroll(C) : course(C,S,H) } 6", literal.toString());
	}

	@Test
	public void testModifyLiteral_shouldIncrementLowerBound_2()
			throws Exception {
		String code = "0 { a(C) : b(C) } 1.";
		List<LineBase> program = parser.parse(code);
		AggregateLiteral literal = (AggregateLiteral) ((GringoRule) program
				.get(0)).getHead().get(0);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextBoolean()).thenReturn(true).thenReturn(true);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(literal);
		assertEquals("1 { a(C) : b(C) } 1", literal.toString());
	}

	@Test
	public void testModifyLiteral_shouldIncrementLowerBound_3()
			throws Exception {
		String code = "0 { enroll(C) : course(C,S,H) } 1.";
		List<LineBase> program = parser.parse(code);
		AggregateLiteral literal = (AggregateLiteral) ((GringoRule) program
				.get(0)).getHead().get(0);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextBoolean()).thenReturn(true).thenReturn(false);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(literal);
		assertEquals("1 { enroll(C) : course(C,S,H) } 1", literal.toString());
	}

	@Test
	public void testModifyLiteral_shouldIncrementUpperBound_1()
			throws Exception {
		String code = "3 { enroll(C) : course(C,S,H) } 4.";
		List<LineBase> program = parser.parse(code);
		AggregateLiteral literal = (AggregateLiteral) ((GringoRule) program
				.get(0)).getHead().get(0);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextBoolean()).thenReturn(false).thenReturn(true);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(literal);
		assertEquals("3 { enroll(C) : course(C,S,H) } 5", literal.toString());
	}

	@Test
	public void testModifyLiteral_shouldIncrementUpperBound_2()
			throws Exception {
		String code = "1 { enroll(C) : course(C,1) } 1.";
		List<LineBase> program = parser.parse(code);
		AggregateLiteral literal = (AggregateLiteral) ((GringoRule) program
				.get(0)).getHead().get(0);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextBoolean()).thenReturn(false).thenReturn(true);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(literal);
		assertEquals("1 { enroll(C) : course(C,1) } 2", literal.toString());
	}

	@Test
	public void testModifyLiteral_shouldIncrementUpperBound_3()
			throws Exception {
		String code = "0 { enroll(C) : course(C,S,H) } 0.";
		List<LineBase> program = parser.parse(code);
		AggregateLiteral literal = (AggregateLiteral) ((GringoRule) program
				.get(0)).getHead().get(0);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextBoolean()).thenReturn(true).thenReturn(true);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(literal);
		assertEquals("0 { enroll(C) : course(C,S,H) } 1", literal.toString());
	}

	@Test
	public void testModifyLiteral_shouldIncrementUpperBound_5()
			throws Exception {
		String code = "2 { enroll(C) : course(C,S,H) } 2.";
		List<LineBase> program = parser.parse(code);
		AggregateLiteral literal = (AggregateLiteral) ((GringoRule) program
				.get(0)).getHead().get(0);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextBoolean()).thenReturn(false).thenReturn(false);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(literal);
		assertEquals("2 { enroll(C) : course(C,S,H) } 3", literal.toString());
	}

	@Test
	public void testModifyLiteral_shouldIncrementUpperBound_6()
			throws Exception {
		String code = "{ enroll(C) : course(C,S,H) } 2.";
		List<LineBase> program = parser.parse(code);
		AggregateLiteral literal = (AggregateLiteral) ((GringoRule) program
				.get(0)).getHead().get(0);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextBoolean()).thenReturn(true).thenReturn(true);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(literal);
		assertEquals("{ enroll(C) : course(C,S,H) } 3", literal.toString());
	}

}
