/*******************************************************************************
 * Copyright (c) 2011 Martin Schwengerer.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Martin Schwengerer - initial API and implementation
 ******************************************************************************/
package aspmodificator.modificationrules.gringo;

import static org.junit.Assert.*;

import java.util.List;
import java.util.Random;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import org.junit.Before;
import org.junit.Test;

import aspmodificator.model.CommentLine;
import aspmodificator.model.LineBase;
import aspmodificator.parser.GringoParser;
import aspmodificator.parser.gringo.GringoRule;
import static org.hamcrest.CoreMatchers.*;
import static org.hamcrest.core.Is.is;

public class DeleteRuleModificationRuleTest {
	private GringoParser parser;
	private DeleteRuleModificationRule rule;

	@Before
	public void setUp() throws Exception {
		rule = new DeleteRuleModificationRule();
		parser = new GringoParser();
	}

	@Test
	public void testApply_shouldDeleteConstraint_1() throws Exception {
		String code = ":- X != 1, b(X), not c(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		rule.setRandom(mockedRandom);
		rule.setRuleType(LogicRuleTypeEnum.CONSTRAINT);

		int res = rule.apply(program);

		assertThat(res, is(1));

		assertThat(program, is(notNullValue()));
		assertThat(program.size(), is(0));
	}

	@Test
	public void testApply_shouldDeleteConstraint_2() throws Exception {
		String code = "a(1). \n :- X != 1, b(X), not c(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		rule.setRandom(mockedRandom);
		rule.setRuleType(LogicRuleTypeEnum.CONSTRAINT);

		int res = rule.apply(program);

		assertThat(res, is(1));

		assertThat(program, is(notNullValue()));
		assertThat(program.size(), is(1));
		assertThat(program.get(0).toString(), is("a(1)."));
	}

	@Test
	public void testApply_shouldDeleteConstraint_3() throws Exception {
		String code = "good(x-man). :- good(X),evil(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		rule.setRandom(mockedRandom);
		rule.setRuleType(LogicRuleTypeEnum.CONSTRAINT);

		int res = rule.apply(program);

		assertThat(res, is(1));

		assertThat(program, is(notNullValue()));
		assertThat(program.size(), is(1));
		assertThat(program.get(0).toString(), is("good(x-man)."));
	}

	@Test
	public void testApply_shouldDeleteFact_1() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).\n" + "fact.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		rule.setRandom(mockedRandom);
		rule.setRuleType(LogicRuleTypeEnum.FACT);

		int res = rule.apply(program);

		assertThat(res, is(1));

		assertThat(program, is(notNullValue()));
		assertThat(program.size(), is(1));
		assertThat(program.get(0).toString(),
				is("a(X) :- X != 1, b(X), not c(X)."));
	}

	@Test
	public void testApply_shouldDeleteFact_2() throws Exception {
		String code = "good(x-man). :- good(X),evil(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		rule.setRandom(mockedRandom);
		rule.setRuleType(LogicRuleTypeEnum.FACT);

		int res = rule.apply(program);

		assertThat(res, is(1));

		assertThat(program, is(notNullValue()));
		assertThat(program.size(), is(1));
		assertThat(program.get(0).toString(), is(":- good(X), evil(X)."));
	}

	@Test
	public void testApply_shouldDeleteNothing() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setRandom(new Random(0));
		rule.setMinApplication(0);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(0, res);

		assertNotNull(program);
		assertEquals(1, program.size());
		assertEquals(GringoRule.class, program.get(0).getClass());
	}

	@Test
	public void testApply_shouldDeleteNothing_1() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		rule.setRandom(mockedRandom);
		rule.setRuleType(LogicRuleTypeEnum.FACT);

		int res = rule.apply(program);

		assertThat(res, is(0));

		assertThat(program, is(notNullValue()));
		assertThat(program.size(), is(1));
		assertThat(program.get(0).toString(),
				is("a(X) :- X != 1, b(X), not c(X)."));
	}

	@Test
	public void testApply_shouldDeleteNothing_2() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		rule.setRandom(mockedRandom);
		rule.setRuleType(LogicRuleTypeEnum.CONSTRAINT);

		int res = rule.apply(program);

		assertThat(res, is(0));

		assertThat(program, is(notNullValue()));
		assertThat(program.size(), is(1));
		assertThat(program.get(0).toString(),
				is("a(X) :- X != 1, b(X), not c(X)."));
	}

	@Test
	public void testApply_shouldDeleteNothing_3() throws Exception {
		String code = "fact.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		rule.setRandom(mockedRandom);
		rule.setRuleType(LogicRuleTypeEnum.PROPER);

		int res = rule.apply(program);

		assertThat(res, is(0));

		assertThat(program, is(notNullValue()));
		assertThat(program.size(), is(1));
		assertThat(program.get(0).toString(), is("fact."));
	}

	@Test
	public void testApply_shouldDeleteOneRule_1() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(true);
		rule.setRandom(new Random(1));
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(1, program.size());
		assertEquals(CommentLine.class, program.get(0).getClass());
	}

	@Test
	public void testApply_shouldDeleteOneRule_2() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setRandom(new Random(1));
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(0, program.size());
	}

	@Test
	public void testApply_shouldDeleteOneRule_3() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).\n%*Hallo*%";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setRandom(new Random(1));
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(1, program.size());
		assertEquals(CommentLine.class, program.get(0).getClass());
	}

	@Test
	public void testApply_shouldDeleteOneRule_4() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).\nb(X) :- a(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setRandom(new Random(1));
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(1, program.size());
		assertEquals(GringoRule.class, program.get(0).getClass());
	}

	@Test
	public void testApply_shouldDeleteOneRule_5() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).\n" + "b(X) :- a(X).\n"
				+ "c(Y) :- a(Y), not b(Y).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(true);

		rule.setRandom(new Random(0));

		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(3, program.size());
		assertEquals(GringoRule.class, program.get(0).getClass());
		assertEquals(CommentLine.class, program.get(1).getClass());
		assertEquals(GringoRule.class, program.get(2).getClass());
	}

	@Test
	public void testApply_shouldDeleteOneRule_6() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setRandom(new Random(99999999));
		rule.setMinApplication(0);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(0, program.size());
	}

	@Test
	public void testApply_shouldDeleteProperRule_1() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).\n fact.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		rule.setRandom(mockedRandom);
		rule.setRuleType(LogicRuleTypeEnum.PROPER);

		int res = rule.apply(program);

		assertThat(res, is(1));

		assertThat(program, is(notNullValue()));
		assertThat(program.size(), is(1));
		assertThat(program.get(0).toString(), is("fact."));
	}

	@Test
	public void testApply_shouldDeleteProperRule_2() throws Exception {
		String code = "p(X) :- q(X). q(X) :- r(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(1.0).thenReturn(0.0);
		rule.setRandom(mockedRandom);
		rule.setRuleType(LogicRuleTypeEnum.PROPER);

		int res = rule.apply(program);

		assertThat(res, is(1));

		assertThat(program, is(notNullValue()));
		assertThat(program.size(), is(1));
		assertThat(program.get(0).toString(), is("p(X) :- q(X)."));
	}

	@Test
	public void testApply_shouldDeleteTwoRule_1() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).\n" + "b(X) :- a(X).\n"
				+ "c(Y) :- a(Y), not b(Y).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setRandom(new Random(1));
		rule.setMinApplication(2);
		rule.setMaxApplication(2);

		int res = rule.apply(program);

		assertEquals(2, res);

		assertNotNull(program);
		assertEquals(1, program.size());
		assertEquals(GringoRule.class, program.get(0).getClass());
	}

	@Test
	public void testApply_shouldThrowException() throws Exception {
		String code = "%Hallo\n%Only Comments";
		List<LineBase> program = parser.parse(code);

		assertEquals(2, program.size());

		rule.setWriteComments(false);
		rule.setRandom(new Random(1));
		rule.setMinApplication(2);
		rule.setMaxApplication(2);

		int res = rule.apply(program);
		assertEquals(0, res);
	}

}
