/*******************************************************************************
 * Copyright (c) 2011 Martin Schwengerer.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Martin Schwengerer - initial API and implementation
 ******************************************************************************/
package aspmodificator.modificationrules.gringo;

import static org.hamcrest.core.Is.is;
import static org.junit.Assert.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.util.List;
import java.util.Random;

import org.junit.Before;
import org.junit.Test;

import aspmodificator.model.LineBase;
import aspmodificator.modificationrules.TabuList;
import aspmodificator.parser.GringoParser;
import aspmodificator.parser.gringo.GringoRule;
import aspmodificator.parser.gringo.Literal;
import aspmodificator.parser.gringo.MathTerm;

public class MathOperatorsModificationRuleTest {
	private GringoParser parser;
	private MathOperatorsModificationRule rule;

	@Before
	public void setUp() throws Exception {
		rule = new MathOperatorsModificationRule();
		rule.setTabuList(new TabuList());
		parser = new GringoParser();
	}

	@Test
	public void testApply_shouldDoNothing() throws Exception {
		String code = "a(X) :- b(X), c(X).";
		List<LineBase> program = parser.parse(code);

		String original = program.get(0).toString();

		Random mockedRandom = mock(Random.class);
		rule.setRandom(mockedRandom);

		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);
		assertEquals(0, res);

		assertEquals(original, program.get(0).toString());
	}

	@Test
	public void testApply_shouldModify_1() throws Exception {
		String code = "a(X,Y,Z) :- Z = X + Y.";
		List<LineBase> program = parser.parse(code);

		Random mockedRandom = mock(Random.class);
		when(
				mockedRandom
						.nextInt(MathOperatorsModificationRule.dualOperators.length))
				.thenReturn(0);
		rule.setRandom(mockedRandom);

		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Literal lit = ((GringoRule) program.get(0)).getBody().get(0);
		MathTerm mathTerm = (MathTerm) lit.getTerms().get(1);

		assertEquals(MathTerm.PLUS, mathTerm.getIdentifier());

		int res = rule.apply(program);
		assertEquals(1, res);

		assertEquals(MathTerm.DIV, mathTerm.getIdentifier());
	}

	@Test
	public void testApply_shouldModify_2() throws Exception {
		String code = "a(X,Y,Z) :- Z = X + Y, num(X), num(Y), num(Z).";
		List<LineBase> program = parser.parse(code);

		Random mockedRandom = mock(Random.class);
		when(
				mockedRandom
						.nextInt(MathOperatorsModificationRule.dualOperators.length))
				.thenReturn(5);
		rule.setRandom(mockedRandom);

		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Literal lit = ((GringoRule) program.get(0)).getBody().get(0);
		MathTerm mathTerm = (MathTerm) lit.getTerms().get(1);

		assertEquals(MathTerm.PLUS, mathTerm.getIdentifier());

		int res = rule.apply(program);
		assertEquals(1, res);

		assertEquals(MathTerm.MULT, mathTerm.getIdentifier());
	}

	@Test
	public void testApply_shouldModify_3() throws Exception {
		String code = "a(Z) :- a(X), a(Y), Z1 = X - Z2, Z2 = X * Y.";
		List<LineBase> program = parser.parse(code);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(1.0).thenReturn(0.0);
		when(
				mockedRandom
						.nextInt(MathOperatorsModificationRule.dualOperators.length))
				.thenReturn(2);
		rule.setRandom(mockedRandom);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Literal lit = ((GringoRule) program.get(0)).getBody().get(3);
		MathTerm mathTerm = (MathTerm) lit.getTerms().get(1);

		assertEquals(MathTerm.MULT, mathTerm.getIdentifier());

		int res = rule.apply(program);
		assertEquals(1, res);

		assertEquals(MathTerm.PLUS, mathTerm.getIdentifier());
	}

	@Test
	public void testApply_shouldModify_4() throws Exception {
		String code = "next(X,Y) :- r(X;Y),Y=X+1.";
		List<LineBase> program = parser.parse(code);

		Random mockedRandom = mock(Random.class);
		when(
				mockedRandom
						.nextInt(MathOperatorsModificationRule.dualOperators.length))
				.thenReturn(5);
		rule.setRandom(mockedRandom);
		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);
		assertEquals(1, res);

		assertThat(res, is(1));
		assertThat(((GringoRule) program.get(0)).toString(),
				is("next(X,Y) :- r(X;Y), Y = X*1."));
	}

	@Test
	public void testFindApplicableObjects_shouldReturn1_1() throws Exception {
		String code = "a(X,Y,Z) :- X = Y - Z.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(1, literalList.size());
		assertEquals("Y-Z", literalList.get(0).toString());
	}

	@Test
	public void testFindApplicableObjects_shouldReturn1_2() throws Exception {
		String code = "a(X,Y,Z) :- X = Y / Z.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(1, literalList.size());
		assertEquals("Y/Z", literalList.get(0).toString());
	}

	@Test
	public void testFindApplicableObjects_shouldReturn1_3() throws Exception {
		String code = "a(X,Y,Z) :- X =  Y \\ Z.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(1, literalList.size());
		assertEquals("Y\\Z", literalList.get(0).toString());
	}

	@Test
	public void testFindApplicableObjects_shouldReturn1_4() throws Exception {
		String code = "a(X,Y,Z) :- X = Y + Z.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(1, literalList.size());
		assertEquals("Y+Z", literalList.get(0).toString());
	}

	@Test
	public void testFindApplicableObjects_shouldReturn1_5() throws Exception {
		String code = "a(X,Y,Z) :- X = Y ** Z.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(1, literalList.size());
		assertEquals("Y**Z", literalList.get(0).toString());
	}

	@Test
	public void testFindApplicableObjects_shouldReturn1_6() throws Exception {
		String code = "a(X,Y,Z) :- X = Y * Z.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(1, literalList.size());
		assertEquals("Y*Z", literalList.get(0).toString());
	}

	@Test
	public void testFindApplicableObjects_shouldReturn1_7() throws Exception {
		String code = "a(X) :- X = 1 + 3.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(1, literalList.size());
		assertEquals("1+3", literalList.get(0).toString());
	}

	@Test
	public void testFindApplicableObjects_shouldReturn1_8() throws Exception {
		String code = "a(X,Y) :- Y = X - 1.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(1, literalList.size());
		assertEquals("X-1", literalList.get(0).toString());
	}

	@Test
	public void testFindApplicableObjects_shouldReturn2_1() throws Exception {
		String code = "a(X,Y) :- Y = X - T1, T1 = 2 * 2.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(2, literalList.size());
	}

	@Test
	public void testFindApplicableObjects_shouldReturn3() throws Exception {
		String code = "a(X,Y) :- Y = X - T1, T1 = 2 * 2.\n"
				+ "b(X,Y) :- Y = X \\ 3.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(3, literalList.size());
	}

	@Test
	public void testFindApplicableObjects_shouldReturnEmptyList_1()
			throws Exception {
		String code = "";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(0, literalList.size());
	}

	@Test
	public void testFindApplicableObjects_shouldReturnEmptyList_2()
			throws Exception {
		String code = "a(x). \n a(X) :- b(X). \n :- c(x).\n a(X) :- X == 1.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(0, literalList.size());
	}

	@Test
	public void testModifyLiteral_shouldModify_1() throws Exception {
		String code = "a(X,Y,Z) :- X = Y / Z.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);
		Literal lit = literalList.get(0);

		rule.modifyLiteral(lit, MathOperatorsModificationRule.dualOperators[1]);

		assertEquals(MathOperatorsModificationRule.dualOperators[1],
				lit.getIdentifier());
	}

	@Test
	public void testModifyLiteral_shouldModify_2() throws Exception {
		String code = "a(X,Y,Z) :- X = Y - Z.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);
		Literal lit = literalList.get(0);

		Random mockedRandom = mock(Random.class);
		when(
				mockedRandom
						.nextInt(RelationOperatorModificationRule.operators.length))
				.thenReturn(0);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(lit, MathOperatorsModificationRule.dualOperators[0]);

		assertEquals(MathOperatorsModificationRule.dualOperators[0],
				lit.getIdentifier());
	}

	@Test
	public void testModifyLiteral_shouldModify_3() throws Exception {
		String code = "a(X,Y,Z) :- X = Y / Z.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);
		Literal lit = literalList.get(0);

		Random mockedRandom = mock(Random.class);
		when(
				mockedRandom
						.nextInt(RelationOperatorModificationRule.operators.length))
				.thenReturn(0).thenReturn(3);
		rule.setRandom(mockedRandom);

		rule.modifyLiteral(lit, MathOperatorsModificationRule.dualOperators[3]);

		assertEquals(MathOperatorsModificationRule.dualOperators[3],
				lit.getIdentifier());
	}

	@Test
	public void testTabuList_shouldCreateDifferentMutants_1() throws Exception {
		String code = "next(X,Y) :- r(X;Y),Y=X+1.";
		List<LineBase> program = parser.parse(code);

		Random mockedRandom = mock(Random.class);
		when(
				mockedRandom
						.nextInt(MathOperatorsModificationRule.dualOperators.length))
				.thenReturn(5).thenReturn(5).thenReturn(0);
		rule.setRandom(mockedRandom);
		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setUseMemory(true);

		int res = rule.apply(program);

		assertThat(res, is(1));
		assertThat(((GringoRule) program.get(0)).toString(),
				is("next(X,Y) :- r(X;Y), Y = X*1."));

		// second mutant
		program = parser.parse(code);
		assertThat(((GringoRule) program.get(0)).toString(),
				is("next(X,Y) :- r(X;Y), Y = X+1."));
		res = rule.apply(program);

		assertThat(res, is(1));
		assertThat(((GringoRule) program.get(0)).toString(),
				is("next(X,Y) :- r(X;Y), Y = X/1."));
	}

	@Test
	public void testTabuList_shouldCreateDifferentMutants_2() throws Exception {
		String code = "next(X,Y) :- r(X;Y),Y=X+1, X=Y-1.";
		List<LineBase> program = parser.parse(code);

		Random mockedRandom = mock(Random.class);
		when(
				mockedRandom
						.nextInt(MathOperatorsModificationRule.dualOperators.length))
				.thenReturn(5).thenReturn(5).thenReturn(0).thenReturn(5);
		when(mockedRandom.nextDouble()).thenReturn(0.0).thenReturn(0.0)
				.thenReturn(1.0).thenReturn(0.0);
		rule.setRandom(mockedRandom);
		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setUseMemory(true);

		int res = rule.apply(program);

		assertThat(res, is(1));
		assertThat(((GringoRule) program.get(0)).toString(),
				is("next(X,Y) :- r(X;Y), Y = X*1, X = Y-1."));

		// second mutant
		program = parser.parse(code);
		assertThat(((GringoRule) program.get(0)).toString(),
				is("next(X,Y) :- r(X;Y), Y = X+1, X = Y-1."));
		res = rule.apply(program);

		assertThat(res, is(1));
		assertThat(((GringoRule) program.get(0)).toString(),
				is("next(X,Y) :- r(X;Y), Y = X/1, X = Y-1."));

		// third mutant
		program = parser.parse(code);
		assertThat(((GringoRule) program.get(0)).toString(),
				is("next(X,Y) :- r(X;Y), Y = X+1, X = Y-1."));
		res = rule.apply(program);

		assertThat(res, is(1));
		assertThat(((GringoRule) program.get(0)).toString(),
				is("next(X,Y) :- r(X;Y), Y = X+1, X = Y*1."));
	}

	@Test
	public void testTabuList_shouldCreateSimilarMutants_1() throws Exception {
		String code = "next(X,Y) :- r(X;Y),Y=X+1.";
		List<LineBase> program = parser.parse(code);

		Random mockedRandom = mock(Random.class);
		when(
				mockedRandom
						.nextInt(MathOperatorsModificationRule.dualOperators.length))
				.thenReturn(5).thenReturn(5).thenReturn(0);
		rule.setRandom(mockedRandom);
		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setUseMemory(false);

		int res = rule.apply(program);

		assertThat(res, is(1));
		assertThat(((GringoRule) program.get(0)).toString(),
				is("next(X,Y) :- r(X;Y), Y = X*1."));

		// second mutant
		program = parser.parse(code);

		res = rule.apply(program);

		assertThat(res, is(1));
		assertThat(((GringoRule) program.get(0)).toString(),
				is("next(X,Y) :- r(X;Y), Y = X*1."));
	}

}
