/*******************************************************************************
 * Copyright (c) 2011 Martin Schwengerer.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Martin Schwengerer - initial API and implementation
 ******************************************************************************/
package aspmodificator.modificationrules.gringo;

import static org.hamcrest.CoreMatchers.notNullValue;
import static org.hamcrest.core.Is.is;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertThat;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.util.List;
import java.util.Random;

import org.junit.Before;
import org.junit.Test;

import aspmodificator.model.LineBase;
import aspmodificator.parser.GringoParser;
import aspmodificator.parser.gringo.GringoRule;
import aspmodificator.parser.gringo.Literal;

import common.Tripple;

public class RenameVariableModificationRuleTest {

	private GringoParser parser;
	private RenameVariableModificationRule rule;

	@Before
	public void setUp() throws Exception {
		rule = new RenameVariableModificationRule();
		parser = new GringoParser();
	}

	@Test
	public void testApply_shouldModify_1() throws Exception {
		String code = "a(X,Y) :- b(X), b(Y).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setUseMemory(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextInt(1)).thenReturn(0);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertThat(res, is(1));
		assertThat(((GringoRule) program.get(0)).getHead().get(0).toString(),
				is("a(Y,Y)"));
	}

	@Test
	public void testApply_shouldModify_2() throws Exception {
		String code = "a(X,Y) :- b(X), b(Y).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setUseMemory(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(1.0).thenReturn(0.0);
		when(mockedRandom.nextInt(1)).thenReturn(0);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertThat(res, is(1));

		assertThat(((GringoRule) program.get(0)).getHead().get(0).toString(),
				is("a(X,X)"));
	}

	@Test
	public void testApply_shouldModify_3() throws Exception {
		String code = "first(X) :- rel(X,Y).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setUseMemory(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextInt(1)).thenReturn(0);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertThat(res, is(1));
		assertThat(((GringoRule) program.get(0)).getHead().get(0).toString(),
				is("first(Y)"));

	}

	@Test
	public void testApply_shouldModify_4() throws Exception {
		String code = "first(X) :- rel(X,Y,Z).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setUseMemory(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextInt(2)).thenReturn(1);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertThat(res, is(1));
		assertEquals("first(Z)", ((GringoRule) program.get(0)).getHead().get(0)
				.toString());

	}

	@Test
	public void testApply_shouldModify_5() throws Exception {
		String code = ":- sum(S), panDiaUp(_,_,O), S [ sqrHelper3(R,C,A) = A ] S.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setUseMemory(false);
		
		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(1.0).thenReturn(0.0);
		when(mockedRandom.nextInt(3)).thenReturn(1);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertThat(res, is(1));
		assertEquals("panDiaUp(S,_,O)", ((GringoRule) program.get(0))
				.getBody().get(1).toString());

	}

	@Test
	public void testFindApplicableObjects_shouldReturn0_1() throws Exception {
		String code = "a :- b(X).";
		List<LineBase> program = parser.parse(code);
		List<Tripple<Literal, Literal, List<String>>> literalList = rule
				.findApplicableObjects(program);

		assertThat(literalList, is(notNullValue()));
		assertThat(literalList.size(), is(0));
	}

	@Test
	public void testFindApplicableObjects_shouldReturn0_2() throws Exception {
		String code = "a(X) :- b(X).";
		List<LineBase> program = parser.parse(code);
		List<Tripple<Literal, Literal, List<String>>> literalList = rule
				.findApplicableObjects(program);

		assertThat(literalList, is(notNullValue()));
		assertThat(literalList.size(), is(0));
	}

	@Test
	public void testFindApplicableObjects_shouldReturn1_1() throws Exception {
		String code = "a(X) :- b(Y), X = Y + 1.";
		List<LineBase> program = parser.parse(code);
		List<Tripple<Literal, Literal, List<String>>> literalList = rule
				.findApplicableObjects(program);

		assertThat(literalList, is(notNullValue()));
		assertThat(literalList.size(), is(1));
	}

	@Test
	public void testFindApplicableObjects_shouldReturn2_1() throws Exception {
		String code = "a(X,Y) :- b(X), b(Y).";
		List<LineBase> program = parser.parse(code);
		List<Tripple<Literal, Literal, List<String>>> literalList = rule
				.findApplicableObjects(program);

		assertThat(literalList, is(notNullValue()));
		assertThat(literalList.size(), is(2));
	}

	@Test
	public void testFindApplicableObjects_shouldReturn2_2() throws Exception {
		String code = "a :- b(X), b(Y), X != Y.";
		List<LineBase> program = parser.parse(code);
		List<Tripple<Literal, Literal, List<String>>> literalList = rule
				.findApplicableObjects(program);

		assertThat(literalList, is(notNullValue()));
		assertThat(literalList.size(), is(2));
	}

	@Test
	public void testFindApplicableObjects_shouldReturn3_1() throws Exception {
		String code = "a(X) :- b(X), b(Y), X < Y.";
		List<LineBase> program = parser.parse(code);
		List<Tripple<Literal, Literal, List<String>>> literalList = rule
				.findApplicableObjects(program);

		assertThat(literalList, is(notNullValue()));
		assertThat(literalList.size(), is(3));
	}

}
