/*******************************************************************************
 * Copyright (c) 2011 Martin Schwengerer.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Martin Schwengerer - initial API and implementation
 ******************************************************************************/
package aspmodificator.modificationrules.gringo;

import static org.junit.Assert.*;

import java.util.List;
import java.util.Random;

import org.junit.Before;
import org.junit.Test;

import aspmodificator.model.LineBase;
import aspmodificator.parser.GringoParser;
import aspmodificator.parser.gringo.GringoRule;
import aspmodificator.parser.gringo.Literal;

public class SwitchDefaultNegationModificationRuleTest {

	private GringoParser parser;
	private SwitchDefaultNegationModificationRule rule;

	@Before
	public void setUp() throws Exception {
		parser = new GringoParser();
		rule = new SwitchDefaultNegationModificationRule();
	}

	@Test
	public void testAddComments_shouldAddComment_1() throws Exception {
		String code = "a(X) :- X != 1, b(X), c(X).";
		List<LineBase> program = parser.parse(code);
		Literal lit = ((GringoRule) program.get(0)).getBody().get(2); // not
																		// c(X)
		assertFalse(lit.isSigned());
		rule.modifyLiteral(lit);

		assertTrue(lit.isSigned());

		assertNotNull(lit.getSucceedingComments());
		assertEquals(1, lit.getSucceedingComments().size());
		assertEquals(SwitchDefaultNegationModificationRule.COMMENT_ADDED, lit
				.getSucceedingComments().get(0).toString());
	}

	@Test
	public void testAddComments_shouldAddComment_2() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).";
		List<LineBase> program = parser.parse(code);
		Literal lit = ((GringoRule) program.get(0)).getBody().get(2); // not
																		// c(X)

		rule.setWriteComments(true);
		assertTrue(lit.isSigned());
		assertNull(lit.getSucceedingComments());

		rule.modifyLiteral(lit);

		assertFalse(lit.isSigned());

		assertNotNull(lit.getSucceedingComments());
		assertEquals(1, lit.getSucceedingComments().size());
		assertEquals(SwitchDefaultNegationModificationRule.COMMENT_REMOVED, lit
				.getSucceedingComments().get(0).toString());
	}

	@Test
	public void testAddComments_shouldNotAddComment() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).";
		List<LineBase> program = parser.parse(code);
		Literal lit = ((GringoRule) program.get(0)).getBody().get(2); // not
																		// c(X)

		rule.setWriteComments(false);
		assertTrue(lit.isSigned());
		assertNull(lit.getSucceedingComments());

		rule.modifyLiteral(lit);

		assertFalse(lit.isSigned());

		assertNull(lit.getSucceedingComments());
	}

	@Test
	public void testApply_shouldSwitchSigned_1() throws Exception {
		String code = "a(X) :- X != 1, b(X), c(X).";
		List<LineBase> program = parser.parse(code);

		rule.setRandom(new Random(1));
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		Literal lit = ((GringoRule) program.get(0)).getBody().get(2); // c(X)
		assertFalse(lit.isSigned());

		int res = rule.apply(program);
		assertEquals(1, res);

		assertTrue(lit.isSigned());
	}

	@Test
	public void testApply_shouldSwitchSigned_2() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).";
		List<LineBase> program = parser.parse(code);

		rule.setRandom(new Random(1));
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		Literal lit = ((GringoRule) program.get(0)).getBody().get(2); // not
																		// c(X)
		assertTrue(lit.isSigned());

		int res = rule.apply(program);
		assertEquals(1, res);

		assertFalse(lit.isSigned());
	}

	@Test
	public void testFindApplicableObjects_shouldReturn1() throws Exception {
		String code = "a(x) :- b(x).";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(1, literalList.size());
		assertEquals("b(x)", literalList.get(0).toString());
	}

	@Test
	public void testFindApplicableObjects_shouldReturnEmptyList()
			throws Exception {
		String code = "";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(0, literalList.size());
	}

	@Test
	public void testFindApplicableObjects_shouldReturnEmptyList_2()
			throws Exception {
		String code = "a(X,Y,Z) :- Z = X+Y.";
		List<LineBase> program = parser.parse(code);
		List<Literal> literalList = rule.findApplicableObjects(program);

		assertNotNull(literalList);
		assertEquals(0, literalList.size());
	}

	@Test
	public void testModifyLiteral_shouldSwitchSigned_1() throws Exception {
		String code = "a(X) :- X != 1, b(X), not c(X).";
		List<LineBase> program = parser.parse(code);
		Literal lit = ((GringoRule) program.get(0)).getBody().get(2); // not
																		// c(X)

		rule.setWriteComments(true);
		assertTrue(lit.isSigned());

		rule.modifyLiteral(lit);

		assertFalse(lit.isSigned());
	}

	@Test
	public void testModifyLiteral_shouldSwitchSigned_2() throws Exception {
		String code = "a(X) :- X != 1, b(X), c(X).";
		List<LineBase> program = parser.parse(code);
		Literal lit = ((GringoRule) program.get(0)).getBody().get(2); // not
																		// c(X)
		assertFalse(lit.isSigned());
		rule.modifyLiteral(lit);

		assertTrue(lit.isSigned());

	}
}
