/*******************************************************************************
 * Copyright (c) 2011 Martin Schwengerer.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Martin Schwengerer - initial API and implementation
 ******************************************************************************/
package aspmodificator.modificationrules.gringo;

import static org.junit.Assert.assertEquals;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.util.List;
import java.util.Random;

import org.junit.Before;
import org.junit.Test;

import aspmodificator.model.LineBase;
import aspmodificator.parser.GringoParser;
import aspmodificator.parser.gringo.GringoRule;

public class TwiddleAggregateWeightsModificationRuleTest {

	private GringoParser parser;
	private TwiddleAggregateWeightsModificationRule rule;

	@Before
	public void setUp() throws Exception {
		rule = new TwiddleAggregateWeightsModificationRule();
		parser = new GringoParser();
	}

	@Test
	public void testApply_shouldAddComment_1() throws Exception {
		String code = "2 #sum [a, not b, c=2] 3.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(true);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextBoolean()).thenReturn(true);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertEquals(1, res);
		assertEquals(
				"2 #sum [ a, not b, c = 3%*ModificationNote: Weight increased by 1 *% ] 3",
				((GringoRule) program.get(0)).getHead().get(0).toString());
	}

	@Test
	public void testApply_shouldModify_1() throws Exception {
		String code = "2 #sum [a, not b, c=2] 3.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextBoolean()).thenReturn(true);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertEquals(1, res);
		assertEquals("2 #sum [ a, not b, c = 3 ] 3",
				((GringoRule) program.get(0)).getHead().get(0).toString());
	}

	@Test
	public void testApply_shouldModify_2() throws Exception {
		String code = "2 #sum [a = 1, not b, c=2] 3.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(1.0).thenReturn(0.0);
		when(mockedRandom.nextBoolean()).thenReturn(true);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertEquals(1, res);
		assertEquals("2 #sum [ a = 1, not b, c = 3 ] 3",
				((GringoRule) program.get(0)).getHead().get(0).toString());
	}

	@Test
	public void testApply_shouldModify_3() throws Exception {
		String code = "2 #sum [a, not b, c=2] 3.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(1.0).thenReturn(1.0)
				.thenReturn(0.0);
		when(mockedRandom.nextBoolean()).thenReturn(false);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertEquals(1, res);
		assertEquals("2 #sum [ a, not b, c = 1 ] 3",
				((GringoRule) program.get(0)).getHead().get(0).toString());
	}

	@Test
	public void testApply_shouldModify_4() throws Exception {
		String code = "ok :- 2 [val(P,V)=V] 8.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextBoolean()).thenReturn(true);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertEquals(1, res);
		assertEquals("2 [ val(P,V) = V+1 ] 8", ((GringoRule) program.get(0))
				.getBody().get(0).toString());
	}

	@Test
	public void testApply_shouldModify_5() throws Exception {
		String code = "ok :- 2 [val(P,V)=V+1] 8.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextBoolean()).thenReturn(true);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertEquals(1, res);
		assertEquals("2 [ val(P,V) = V+1+1 ] 8", ((GringoRule) program.get(0))
				.getBody().get(0).toString());
	}

	@Test
	public void testApply_shouldNotModify_1() throws Exception {
		String code = "	{ push(B,right,B1,S) } :- reachable(L,S), right(L,B), box_step(B,S), pushable_right(B,B1,S), good_pushlocation(B1), actionstep(S).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextBoolean()).thenReturn(true);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertEquals(0, res);
		assertEquals("{ push(B,right,B1,S) }", ((GringoRule) program.get(0))
				.getHead().get(0).toString());
	}

	@Test
	public void testApply_shouldNotModify_2() throws Exception {
		String code = "	:- { push(B,right,B1,S) }.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextBoolean()).thenReturn(true);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertEquals(0, res);
		assertEquals("{ push(B,right,B1,S) }", ((GringoRule) program.get(0))
				.getBody().get(0).toString());
	}

	@Test
	public void testApply_shouldNotModify_3() throws Exception {
		String code = "	:- N1 = #count{depot(D,K)}, N2 = #count{altdepot(D1,K1)}, N1 != N2.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextBoolean()).thenReturn(true);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertEquals(0, res);
		assertEquals(
				":- N1 = #count { depot(D,K) }, N2 = #count { altdepot(D1,K1) }, N1 != N2.",
				((GringoRule) program.get(0)).toOutputString());
	}

	@Test
	public void testApply_shouldNotModify_4() throws Exception {
		String code = "{ altdepot(R,K) } :- restaurant(R,K).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextBoolean()).thenReturn(true);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);
		assertEquals(0, res);
		assertEquals("{ altdepot(R,K) } :- restaurant(R,K).",
				((GringoRule) program.get(0)).toOutputString());
	}

}
