/*******************************************************************************
 * Copyright (c) 2011 Martin Schwengerer.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Martin Schwengerer - initial API and implementation
 ******************************************************************************/
package aspmodificator.modificationrules.gringo;

import static org.hamcrest.core.Is.is;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertThat;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.util.List;
import java.util.Random;

import org.junit.Before;
import org.junit.Test;

import aspmodificator.model.LineBase;
import aspmodificator.parser.GringoParser;
import aspmodificator.parser.gringo.GringoRule;
import aspmodificator.parser.gringo.Literal;
import aspmodificator.parser.gringo.Term;
import aspmodificator.parser.gringo.VarTerm;
import static org.hamcrest.CoreMatchers.*;

import common.Pair;

public class VariableToConstantModificationRuleTest {
	private VariableToConstantModificationRule rule;
	private GringoParser parser;

	@Before
	public void setUp() throws Exception {
		rule = new VariableToConstantModificationRule();
		parser = new GringoParser();
	}

	@Test
	public void testAddComments_shouldAddComment_1() throws Exception {
		String code = "a(x,y) :- b(x), not c(x), a(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(true);
		rule.setRandom(new Random(1));
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(1, program.size());

		Literal lit = ((GringoRule) program.get(0)).getBody().get(2).getTerms()
				.get(0);

		assertNotNull(lit.getSucceedingComments());
		assertEquals(1, lit.getSucceedingComments().size());
		assertEquals(
				VariableToConstantModificationRule.COMMENT_VARIABLE_TO_CONSTANT,
				lit.getSucceedingComments().get(0).toString());
	}

	@Test
	public void testAddComments_shouldNotAddComment_1() throws Exception {
		String code = "a(x,y) :- b(x), not c(x), a(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setRandom(new Random(1));
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(1, program.size());

		Literal lit = ((GringoRule) program.get(0)).getBody().get(2).getTerms()
				.get(0);

		assertEquals("x", lit.toString());

		assertNull(lit.getSucceedingComments());
	}

	@Test
	public void testApply_shouldDoNothing_1() throws Exception {
		String code = "a(x) :- b(x), not c(x).";
		List<LineBase> program = parser.parse(code);
		String oldString = program.get(0).toString();
		rule.setWriteComments(true);
		rule.setRandom(new Random(1));
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(0, res);

		assertNotNull(program);
		assertEquals(1, program.size());
		assertEquals(oldString, program.get(0).toString());
	}

	@Test
	public void testApply_shouldDoNothing_2() throws Exception {
		String code = "a(x,y) :- b(x), not c(x), a(f(x)).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setRandom(new Random(1));
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(0, res);

		assertNotNull(program);
		assertEquals(1, program.size());

		assertEquals("a(f(x))", ((GringoRule) program.get(0)).getBody().get(2)
				.toString());
	}

	@Test
	public void testApply_shouldNotTransform_1() throws Exception {
		String code = "test(X) :- X = 2.";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(1.0).thenReturn(0.0);

		rule.setRandom(mockedRandom);

		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(1, program.size());

		assertEquals("X = 2", ((GringoRule) program.get(0)).getBody().get(0)
				.toString());

		assertEquals("test(x)", ((GringoRule) program.get(0)).getHead().get(0)
				.toString());
	}

	@Test
	public void testApply_shouldTransform_1() throws Exception {
		String code = "a(X,y) :- b(x), not c(x).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setRandom(new Random(1));
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(1, program.size());

		assertEquals("a(x,y)", ((GringoRule) program.get(0)).getHead().get(0)
				.toString());
	}

	@Test
	public void testApply_shouldTransform_10() throws Exception {
		String code = "solution(left). good_pushlocation(L) :- solution(L).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setExistenceCheckTarget(RuleTargetEnum.ALL);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(1.0).thenReturn(0.0)
				.thenReturn(0.0);
		when(mockedRandom.nextInt(1)).thenReturn(0);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);

		assertThat(res, is(1));

		assertThat(program, is(notNullValue()));
		assertThat(program.size(), is(2));

		assertThat(((GringoRule) program.get(1)).getHead().get(0).toString(),
				is("good_pushlocation(left)"));
	}

	@Test
	public void testApply_shouldTransform_11() throws Exception {
		String code = "a(1). :- test(X), X >= 2, a(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(1.0).thenReturn(0.0);

		rule.setRandom(mockedRandom);

		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setMrRuleTarget(RuleTargetEnum.ALL);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(2, program.size());
		assertEquals("test(X)", ((GringoRule) program.get(1)).getBody().get(0)
				.toString());
		assertEquals("X >= 2", ((GringoRule) program.get(1)).getBody().get(1)
				.toString());

		assertEquals("a(x)", ((GringoRule) program.get(1)).getBody().get(2)
				.toString());
	}

	@Test
	public void testApply_shouldTransform_12() throws Exception {
		String code = "test(a,b,c). :- test(_,_,A), X >= 2, a(X), a(A).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);

		rule.setRandom(mockedRandom);

		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setMrRuleTarget(RuleTargetEnum.ALL);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(2, program.size());
		assertEquals("test(_,_,a)", ((GringoRule) program.get(1)).getBody()
				.get(0).toString());
	}

	@Test
	public void testApply_shouldTransform_2() throws Exception {
		String code = "a(X,y) :- aa(X), b(x), not c(x).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setRandom(new Random(100000000));
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(1, program.size());

		assertEquals("a(x,y)", ((GringoRule) program.get(0)).getHead().get(0)
				.toString());
	}

	@Test
	public void testApply_shouldTransform_3() throws Exception {
		String code = "a(x,y) :- aa(x), -b(X), not c(x).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setRandom(new Random(1));
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(1, program.size());

		assertEquals("-b(x)", ((GringoRule) program.get(0)).getBody().get(1)
				.toString());
	}

	@Test
	public void testApply_shouldTransform_4() throws Exception {
		String code = "a(X,y) :- aa(X), b(X), not c(x).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(1.0).thenReturn(0.0);
		when(mockedRandom.nextInt(1)).thenReturn(0);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(1, program.size());

		assertEquals("a(X,y)", ((GringoRule) program.get(0)).getHead().get(0)
				.toString());
		assertEquals("aa(x)", ((GringoRule) program.get(0)).getBody().get(0)
				.toString());
	}

	@Test
	public void testApply_shouldTransform_5() throws Exception {
		String code = "a(x,y) :- b(x), not -c(x), a(f(X)).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setRandom(new Random(1));
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(1, program.size());

		assertEquals("a(f(x))", ((GringoRule) program.get(0)).getBody().get(2)
				.toString());
	}

	@Test
	public void testApply_shouldTransform_6() throws Exception {
		String code = "p(a;b). fail :- p(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setExistenceCheckTarget(RuleTargetEnum.ALL);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextInt(2)).thenReturn(1);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(2, program.size());

		assertEquals("p(a)", ((GringoRule) program.get(1)).getBody().get(0)
				.toString());
	}

	@Test
	public void testApply_shouldTransform_7() throws Exception {
		String code = "p(a;b). fail :- p(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setExistenceCheckTarget(RuleTargetEnum.NONE);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextInt(2)).thenReturn(1);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(2, program.size());

		assertEquals("p(x)", ((GringoRule) program.get(1)).getBody().get(0)
				.toString());
	}

	@Test
	public void testApply_shouldTransform_8() throws Exception {
		String code = "p(a;b). fail :- p(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setExistenceCheckTarget(RuleTargetEnum.HEAD);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextInt(2)).thenReturn(1);
		when(mockedRandom.nextInt(3)).thenReturn(2);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(2, program.size());

		assertEquals("p(x)", ((GringoRule) program.get(1)).getBody().get(0)
				.toString());
	}

	@Test
	public void testApply_shouldTransform_9() throws Exception {
		String code = "p(a;b). fail :- p(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setMinApplication(1);
		rule.setMaxApplication(1);
		rule.setExistenceCheckTarget(RuleTargetEnum.BODY);

		Random mockedRandom = mock(Random.class);
		when(mockedRandom.nextDouble()).thenReturn(0.0);
		when(mockedRandom.nextInt(2)).thenReturn(1);
		rule.setRandom(mockedRandom);

		int res = rule.apply(program);

		assertThat(res, is(1));

		assertThat(program, is(notNullValue()));
		assertThat(program.size(), is(2));

		assertThat(((GringoRule) program.get(1)).getBody().get(0).toString(),
				is("p(a)"));
	}

	@Test
	public void testApply_shouldTransformTermToVarTerm() throws Exception {
		String code = "a(x,y) :- -b(x), not c(x), a(X).";
		List<LineBase> program = parser.parse(code);

		rule.setWriteComments(false);
		rule.setRandom(new Random(1));
		rule.setMinApplication(1);
		rule.setMaxApplication(1);

		Literal lit = ((GringoRule) program.get(0)).getBody().get(2).getTerms()
				.get(0);
		assertEquals(VarTerm.class, lit.getClass());

		int res = rule.apply(program);

		assertEquals(1, res);

		assertNotNull(program);
		assertEquals(1, program.size());

		lit = ((GringoRule) program.get(0)).getBody().get(2).getTerms().get(0);
		assertEquals(Term.class, lit.getClass());
	}

	@Test
	public void testFindApplicableObjects_shouldReturn1() throws Exception {
		String code = "a(X,y,z).";
		List<LineBase> program = parser.parse(code);
		List<Pair<Literal, Literal>> pairList = rule
				.findApplicableObjects(program);

		assertNotNull(pairList);
		assertEquals(1, pairList.size());
	}

	@Test
	public void testFindApplicableObjects_shouldReturn2() throws Exception {
		String code = "a(X) :- b(X), c(X).";
		List<LineBase> program = parser.parse(code);
		List<Pair<Literal, Literal>> pairList = rule
				.findApplicableObjects(program);

		rule.setExistenceCheckTarget(RuleTargetEnum.NONE);

		assertNotNull(pairList);
		assertEquals(3, pairList.size());
		assertEquals("X", pairList.get(0).getFirst().toString());
		assertEquals("X", pairList.get(1).getFirst().toString());
	}

	@Test
	public void testFindApplicableObjects_shouldReturn2_2() throws Exception {
		String code = "a(X,Y) :- b(z), c(X,Y).";
		List<LineBase> program = parser.parse(code);
		List<Pair<Literal, Literal>> pairList = rule
				.findApplicableObjects(program);

		assertNotNull(pairList);
		assertEquals(2, pairList.size());
		assertEquals("X", pairList.get(0).getFirst().toString());
		assertEquals("Y", pairList.get(1).getFirst().toString());
	}

	@Test
	public void testFindApplicableObjects_shouldReturnEmptyList_1()
			throws Exception {
		String code = "";
		List<LineBase> program = parser.parse(code);
		List<Pair<Literal, Literal>> pairList = rule
				.findApplicableObjects(program);

		assertNotNull(pairList);
		assertEquals(0, pairList.size());
	}

	@Test
	public void testFindApplicableObjects_shouldReturnEmptyList_2()
			throws Exception {
		String code = "a(x) :- b(x).";
		List<LineBase> program = parser.parse(code);
		List<Pair<Literal, Literal>> pairList = rule
				.findApplicableObjects(program);

		assertNotNull(pairList);
		assertEquals(0, pairList.size());
	}
}
